//! Macros were created because functions started to get really complicated.
//! These macros should not be exported out of the library as they are
//! not safe to use since they do no type checking.



/// Handles the testing of a buffers size and returning the correct
/// error if a buffer does not have the correct amount of bytes.
#[macro_export]
macro_rules! check_length
{
   // This tests for the buffers length being less than the min value.
   // [min_bytes, ∞]
   ($buffer: ident >= $min_bytes: expr) =>
   {
      if $buffer.len() < $min_bytes
      {
         return Err(::conversion_error::ConversionError::TooFewBytes);
      }
   };

   // This tests for the buffers length being more than the max value.
   // [0, max_value]
   ($buffer: ident <= $max_bytes: expr) =>
   {
      if $buffer.len() > $max_bytes
      {
         return Err(::conversion_error::ConversionError::TooManyBytes);
      }
   };

   // This tests for the buffers length being within the given boundaries.
   // [min_value, max_value]
   ($buffer: ident, $min_bytes: expr, $max_bytes: expr) =>
   {
      check_length!($buffer >= $min_bytes);
      check_length!($buffer <= $max_bytes);
   };

   // This tests for the buffers length being equal to the given amount.
   // [bytes, bytes]
   ($buffer: ident == $bytes: expr) =>
   {
      check_length!($buffer, $bytes, $bytes);
   }
}

/// Handles the repetative bit packing for big endian.
macro_rules! pack_big_endian
{
   ($buffer: ident, $target_type: ty, $bytes: expr) =>
   {
      {
         let mut value: $target_type;

         check_length!($buffer >= $bytes);

         // Start with a value of zero and OR it with
         // the bits shifted values from the buffer.
         value = 0;
         for i in 0..$bytes
         {
            value |= ($buffer[i] as $target_type) << ((($bytes - i) - 1) * 8);
         }

         Ok(value)
      }
   }
}

/// Handles the repetative bit packing for little endian.
macro_rules! pack_little_endian
{
   ($buffer: ident, $target_type: ty, $bytes: expr) =>
   {
      {
         let mut value: $target_type;

         check_length!($buffer >= $bytes);

         // Start with a value of zero and OR it with
         // the bits shifted values from the buffer.
         value = 0;
         for i in 0..$bytes
         {
            value |= ($buffer[($bytes - 1) - i] as $target_type) <<
                     ((($bytes - i) - 1) * 8);
         }

         Ok(value)
      }
   }
}

/// Handles the repetative big endian unpacking.
macro_rules! unpack_big_endian
{
   ($value: expr, $bytes: expr) =>
   {
      {
         let mut buffer: ::spellbook::components::Array<u8>;

         // Create an array with enough space for this value
         // and then bit shift the value into a buffer of bytes.
         buffer = ::spellbook::components::Array::with_capacity($bytes);
         for i in 0..$bytes
         {
            buffer.push(($value >> (($bytes - i) - 1) * 8) as u8);
         }

         // Return the buffer of bytes that represent this value.
         buffer
      }
   }
}

/// Handles the repetative little endian unpacking.
macro_rules! unpack_little_endian
{
   ($value: expr, $bytes: expr) =>
   {
      {
         let mut buffer: ::spellbook::components::Array<u8>;

         // Create an array with enough space for this value
         // and then bit shift the value into a buffer of bytes.
         buffer = ::spellbook::components::Array::with_capacity($bytes);
         for i in 0..$bytes
         {
            buffer.push(($value >> (i * 8)) as u8);
         }

         // Return the buffer of bytes that represent this value.
         buffer
      }
   }
}

/// Handles swapping from big endian to little endian.
macro_rules! swap_big_to_little_endian
{
   ($value: expr, $target_type: ty, $bytes: expr) =>
   {
      {
         let buffer: ::spellbook::components::Array<u8>;

         // Convert the big endian value to bytes.
         buffer = unpack_big_endian!($value, $bytes);

         // Convert the buffer of bytes to a little endian value.
         pack_little_endian!(buffer, $target_type, $bytes)
      }
   }
}

/// Handles swapping from little endian to big endian.
macro_rules! swap_little_to_big_endian
{
   ($value: expr, $target_type: ty, $bytes: expr) =>
   {
      {
         let buffer: ::spellbook::components::Array<u8>;

         // Convert the little endian value to bytes.
         buffer = unpack_little_endian!($value, $bytes);

         // Convert the buffer of bytes to a big endian value.
         pack_big_endian!(buffer, $target_type, $bytes)
      }
   }
}
